#!/usr/bin/perl

# ARCADEmini converter: b2am
# version 0.9.5 date 2005-10-16
# part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
# written by Stefan Schuermans <1stein@schuermans.info>
# Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html

use strict;

print <<EOF;
ARCADEmini converter: b2am
version 0.9.5 date 2005-10-16
part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
written by Stefan Schuermans <1stein\@schuermans.info>
Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html

EOF

if (@ARGV < 1) {
  print <<EOF;
syntax: b2am [<options>] <blm/bmm/bml-file> [...]

options: +/-s: enable/disable integer-factor scaling
         +/-S: enable/disable floating-point-factor scaling
         +/-a: enable/disable changing of aspect-ratio
         +/-p: print/do not print movie after reading and scaling
         +/-c: enable/disable compression (enabled by default)
         +/-C: enable/disable differential compression (enabled by default)
         --key <key> <action>: assign action to key
         --snes0 <snes-key> <action>: assign action to key of first SNES pad
         --snes1 <snes-key> <action>: assign action to key of second SNES pad
         --loop: play movies in a loop (until end is requested by key press)
         --in <blm/bmm/bml-file>: specify an input file (e.g. if filename starts with '-')
         --blm <blm-file>: specify a blm input file
         --bmm <bmm-file>: specify a bmm input file
         --bml <bml-file>: specify a bml input file
         --amasm <amasm-file>: output ARCADEmini assembler
         --ambin <ambin-file>: output ARCADEmini binary

keys: 1 2 3 4 5 6 7 8

snes-keys: up down left right select start A B X Y L R

actions:  none: delete assigned action
          start: go to the beginning of the first movie
          prev: go to previous movie
          again: play current movie again
          next: go to next movie
          end: go to the end of the last movie     

examples: b2am +sa example.blm --amasm out.amasm
          b2am +Sa example.bmm -a example.bml --ambin out.ambin
          b2am --key 1 start --key 2 prev --key 3 again --key 4 next \
               --key 5 end --loop ex.blm am.bmm ple.bml --ambin out.ambin

EOF
  exit;
}

$| = 1; # enable autoflush



# constants
my $ACTION_NONE = 0;
my $ACTION_START = 1;
my $ACTION_PREV = 2;
my $ACTION_AGAIN = 3;
my $ACTION_NEXT = 4;
my $ACTION_END = 5;



# process parameters
my $arg;
my $flags = "cC";
my @actions;
my ($key, $action);
my $loop = 0;
my ($delays, $frames);
my @movies = ();
my @movie_names = ();
my $output = 0;
my $commands;
my @movies = ();
my $flat_commands;
my $i;

for ($key = 0; $key < 40; $key++) { @actions[$key] = $ACTION_NONE; }

for ($arg = 0; $arg < @ARGV; $arg++) {

  # action for key
  if ($ARGV[$arg] eq '--key') {
    $arg++;
    die "missing key behind '--key'\n" if ($arg >= @ARGV);
    $key = txt2key ($ARGV[$arg]);
    die "invalid key \"".$ARGV[$arg]."\"\n" if ($key < 0);
    $arg++;
    die "missing action behind '--key <key>'\n" if ($arg >= @ARGV);
    $action = txt2action ($ARGV[$arg]);
    die "invalid action \"".$ARGV[$arg]."\"\n" if ($action < 0);
    $actions[$key] = $action;
  }

  # action for snes0 key
  elsif ($ARGV[$arg] eq '--snes0') {
    $arg++;
    die "missing snes-key behind '--snes0'\n" if ($arg >= @ARGV);
    $key = txt2sneskey (0, $ARGV[$arg]);
    die "invalid snes-key \"".$ARGV[$arg]."\"\n" if ($key < 0);
    $arg++;
    die "missing action behind '--snes0 <snes-key>'\n" if ($arg >= @ARGV);
    $action = txt2action ($ARGV[$arg]);
    die "invalid action \"".$ARGV[$arg]."\"\n" if ($action < 0);
    $actions[$key] = $action;
  }

  # action for snes1 key
  elsif ($ARGV[$arg] eq '--snes1') {
    $arg++;
    die "missing snes-key behind '--snes1'\n" if ($arg >= @ARGV);
    $key = txt2sneskey (1, $ARGV[$arg]);
    die "invalid snes-key \"".$ARGV[$arg]."\"\n" if ($key < 0);
    $arg++;
    die "missing action behind '--snes1 <snes-key>'\n" if ($arg >= @ARGV);
    $action = txt2action ($ARGV[$arg]);
    die "invalid action \"".$ARGV[$arg]."\"\n" if ($action < 0);
    $actions[$key] = $action;
  }

  # play movies in a loop
  elsif ($ARGV[$arg] eq '--loop') {
    $i = 0;
    for ($key = 0; $key < 40; $key++) { $i = 1 if ($actions[$key] == $ACTION_END); }
    print "warning: loop requested and no end action set\n" if ($i == 0);
    $loop = 1;
  }

  # explicit input file
  elsif ($ARGV[$arg] eq '--in') {
    $arg++;
    die "missing filename behind '--in'\n" if ($arg >= @ARGV);
    die "no input allowed after first output (place output options at the end)\n" if ($output);
    # read, scale, encode movie
    ($delays, $frames) = read_b ($ARGV[$arg]);
    ($delays, $frames) = scale_movie ($delays, $frames, 20, 26, $flags);
    print_movie ($delays, $frames) if ($flags =~ /p/);
    $commands = enc_movie (@movies + 0, $delays, $frames, $flags, [@actions]);
    push (@movies, $commands);
    push (@movie_names, $ARGV[$arg]);
  }

  # explicit blm input file
  elsif ($ARGV[$arg] eq '--blm') {
    $arg++;
    die "missing filename behind '--blm'\n" if ($arg >= @ARGV);
    die "no input allowed after first output (place output options at the end)\n" if ($output);
    # read, scale, encode movie
    ($delays, $frames) = read_blm ($ARGV[$arg]);
    ($delays, $frames) = scale_movie ($delays, $frames, 20, 26, $flags);
    print_movie ($delays, $frames) if ($flags =~ /p/);
    $commands = enc_movie (@movies + 0, $delays, $frames, $flags, [@actions]);
    push (@movies, $commands);
    push (@movie_names, $ARGV[$arg]);
  }

  # explicit bmm input file
  elsif ($ARGV[$arg] eq '--bmm') {
    $arg++;
    die "missing filename behind '--bmm'\n" if ($arg >= @ARGV);
    die "no input allowed after first output (place output options at the end)\n" if ($output);
    # read, scale, encode movie
    ($delays, $frames) = read_bmm ($ARGV[$arg]);
    ($delays, $frames) = scale_movie ($delays, $frames, 20, 26, $flags);
    print_movie ($delays, $frames) if ($flags =~ /p/);
    $commands = enc_movie (@movies + 0, $delays, $frames, $flags, [@actions]);
    push (@movies, $commands);
    push (@movie_names, $ARGV[$arg]);
  }

  # explicit bml input file
  elsif ($ARGV[$arg] eq '--bml') {
    $arg++;
    die "missing filename behind '--bml'\n" if ($arg >= @ARGV);
    die "no input allowed after first output (place output options at the end)\n" if ($output);
    # read, scale, encode movie
    ($delays, $frames) = read_bml ($ARGV[$arg]);
    ($delays, $frames) = scale_movie ($delays, $frames, 20, 26, $flags);
    print_movie ($delays, $frames) if ($flags =~ /p/);
    $commands = enc_movie (@movies + 0, $delays, $frames, $flags, [@actions]);
    push (@movies, $commands);
    push (@movie_names, $ARGV[$arg]);
  }

  # amasm output file
  elsif ($ARGV[$arg] eq '--amasm') {
    $arg++;
    die "missing filename behind '--amasm'\n" if ($arg >= @ARGV);
    # arrange commands into sectors if not done yet
    if (! $output) {
      $flat_commands = flatten_commands ([@movies], @movies + 0, $loop, [@movie_names]);
      $output = 1;
    }
    # write amasm
    write_amasm ($flat_commands, $ARGV[$arg]);
  }

  # ambin output file
  elsif ($ARGV[$arg] eq '--ambin') {
    $arg++;
    die "missing filename behind '--ambin'\n" if ($arg >= @ARGV);
    # arrange commands into sectors if not done yet
    if (! $output) {
      $flat_commands = flatten_commands ([@movies], @movies + 0, $loop, [@movie_names]);
      $output = 1;
    }
    # write ambin
    write_ambin ($flat_commands, $ARGV[$arg]);
  }

  # add flags
  elsif ($ARGV[$arg] =~ /^\+/) {
    foreach (split (undef, substr ($ARGV[$arg], 1))) {
      if ($_ =~ /^[A-Za-z]$/) {
        $flags =~ s/$_//g;
        $flags .= $_;
      }
    }
  }

  # remove flags
  elsif ($ARGV[$arg] =~ /^-/) {
    foreach (split (undef, substr ($ARGV[$arg], 1))) {
      if ($_ =~ /^[A-Za-z]$/) {
        $flags =~ s/$_//g;
      }
    }
  }

  # input file
  else {
    die "no input allowed after first output (place output options at the end)\n" if ($output);
    # read, scale, encode movie
    ($delays, $frames) = read_b ($ARGV[$arg]);
    ($delays, $frames) = scale_movie ($delays, $frames, 20, 26, $flags);
    print_movie ($delays, $frames) if ($flags =~ /p/);
    $commands = enc_movie (@movies + 0, $delays, $frames, $flags, [@actions]);
    push (@movies, $commands);
    push (@movie_names, $ARGV[$arg]);
  }

} # for ($arg ...


# convert key name to key number
# usage: $key = txt2key ($name)
sub txt2key {
  my $name = shift;
  if ($name =~ /^[1-8]$/) { return $name - 1; }
  else { return -1; }
}

# convert snes key name to snes key number
# usage: $sneskey = txt2sneskey ($snes, $name)
sub txt2sneskey {
  my $snes = shift;
  my $name = shift;
  return -1 if ($snes < 0 || $snes > 1);
  if ($name eq 'up') { return $snes * 16 + 8 + 11; }
  elsif ($name eq 'down') { return $snes * 16 + 8 + 10; }
  elsif ($name eq 'left') { return $snes * 16 + 8 + 9; }
  elsif ($name eq 'right') { return $snes * 16 + 8 + 8; }
  elsif ($name eq 'select') { return $snes * 16 + 8 + 13; }
  elsif ($name eq 'start') { return $snes * 16 + 8 + 12; }
  elsif ($name eq 'A') { return $snes * 16 + 8 + 7; }
  elsif ($name eq 'B') { return $snes * 16 + 8 + 15; }
  elsif ($name eq 'X') { return $snes * 16 + 8 + 6; }
  elsif ($name eq 'Y') { return $snes * 16 + 8 + 14; }
  elsif ($name eq 'L') { return $snes * 16 + 8 + 5; }
  elsif ($name eq 'R') { return $snes * 16 + 8 + 4; }
  else { return -1; }
}

# convert action name to action number
# usage: $action = txt2action ($name)
sub txt2action {
  my $name = shift;
  if ($name eq 'none') { return $ACTION_NONE; }
  elsif ($name eq 'start') { return $ACTION_START; }
  elsif ($name eq 'prev') { return $ACTION_PREV; }
  elsif ($name eq 'again') { return $ACTION_AGAIN; }
  elsif ($name eq 'next') { return $ACTION_NEXT; }
  elsif ($name eq 'end') { return $ACTION_END; }
  else { return -1; }
}



# read a blinken file
# tries to identifiy the file type
# calls the appropriate function to read the file
# usage: ($delays, $frames) = read_b( $file );
sub read_b {
  my $file = shift;
  my $line;
  open (F, $file) or die "read_b: can't open input file \"$file\": $!\n";
  while ($line = <F>) {
    chomp $line;
    chomp $line;
    if ($line =~ / *# *BlinkenLights Movie/i) {
      close F;
      return read_blm ($file);
    }
    if ($line =~ / *# *BlinkenMini Movie/i) {
      close F;
      return read_bmm ($file);
    }
    if ($line =~ /<blm[> ]/i) {
      close F;
      return read_bml ($file);
    }
  }
  close F;
  die "read_b: unknown file format in input file \"$file\"\n";
}

# read a blm file
# usage: ($delays, $frames) = read_blm( $file );
sub read_blm {
  my $file = shift;
  my @delays = ();
  my @frames = ();
  my $line;
  my @values;
  open (F, $file) or die "read_blm: can't open input file \"$file\": $!\n";
  print "reading blm \"$file\" ";
  foreach $line (<F>) {
    chomp $line;
    chomp $line;
    $line =~ s/#.*//;
    next unless $line =~ /\S+/;
    if ($line =~ /^\@(\d+)/) {
      push (@delays, bound ($1, 1, 60000));
      print ".";
      next;
    }
    @values = split(undef,$line);
    $_ = int ($_) ? 15 : 0 foreach (@values);
    push (@{$frames[$#delays]}, [@values]); 
  }
  close F;
  print "\n";
  return ([@delays], [@frames]);
}

# read a bmm file
# usage: ($delays, $frames) = read_bmm( $file );
sub read_bmm {
  my $file = shift;
  my @delays = ();
  my @frames = ();
  my $line;
  my @values;
  open (F, $file) or die "read_bmm: can't open input file \"$file\": $!\n";
  print "reading bmm \"$file\" ";
  foreach $line (<F>) {
    chomp $line;
    chomp $line;
    $line =~ s/#.*//;
    next unless $line =~ /\S+/;
    if ($line =~ /^\@(\d+)/) {
      push (@delays, bound ($1, 1, 60000));
      print ".";
      next;
    }
    @values = split(/[ \t]+/,$line);
    $_ = (abs (substr ($_, 0, 1) eq "0" ? int (oct $_) : int ($_)) & 0xFF) >> 4 foreach (@values);
    push (@{$frames[$#delays]}, [@values]); 
  }
  close F;
  print "\n";
  return ([@delays], [@frames]);
}

# read a bml file
# usage: ($delays, $frames) = read_bml( $file );
sub read_bml {
  my $file = shift;
  my @delays = ();
  my @frames = ();
  my $width = 26;
  my $height = 20;
  my $bits = 4;
  my $channels = 1;
  my $line;
  open (F, $file) or die "read_bml: can't open input file \"$file\": $!\n";
  print "reading bml \"$file\" ";
  foreach $line (<F>) {
    while ($line ne "") {
      if ($line =~ s/^[^<]*<blm([^>]*)>//) {
        my $blm = $1;
        if ($blm =~ /width="?([0-9]*)"?/) {
          $width = bound ($1, 1, 1000);
        }
        if ($blm =~ /height="?([0-9]*)"?/) {
          $height = bound ($1, 1, 1000);
        }
        if ($blm =~ /bits="?([0-9]*)"?/) {
          $bits = bound ($1, 1, 8);
        }
        if ($blm =~ /channels="?([0-9]*)"?/) {
          $channels = bound ($1, 1, 3);
        }
      }
      elsif ($line =~ s/^[^<]*<frame([^>]*)>//) {
        my $frm = $1;
        my $duration = 1000;
        if ($frm =~ /duration="?([0-9]*)"?/) {
          $duration = bound ($1, 1, 60000);
        }
        push (@delays, $duration);
        print ".";
      }
      elsif ($line =~ s/^[^<]*<row>([^<]*)<\/row>//) {
        my $row = $1;
        $row =~ s/[^0-9A-Fa-f]//g;
        my $chrs = int (($bits + 3) / 4);
        my $maxval = ((1 << $bits) - 1) * $channels;
        my @values = ();
        my $i;
        my $v;
        while (length ($row) >= $chrs * $channels)
        {
          $v = 0;
          for ($i = 0; $i < $channels; $i++) {
            $row =~ s/([0-9A-Fa-f]{$chrs})//;
            $v += hex $1;
          }
          push (@values, int ($v * 0xF / $maxval));
        }
        push (@{$frames[$#delays]}, [@values]); 
      }
      elsif ($line =~ s/^[^<]*<[^>]*>//) { }
      else { $line = ""; }
    }
  }
  close F;
  print "\n";
  return ([@delays], [@frames]);
}



# scale a movie
# usage: ($delays, $frames) = scale_movie ($delays, $frames, $new_height, $new_width, $flags)
# $flags: 's': allow integer-factor scaling
#         'S': allow floating-point-factor scaling
#         'a': allow changing of aspect-ratio
sub scale_movie {
  my $delays = shift;
  my $old_frames = shift;
  my $new_height = shift;
  my $new_width = shift;
  my $flags = shift;
  my $scale_int = $flags =~ /s/;
  my $scale_float = $flags =~ /S/;
  my $scale_aspect = $flags =~ /a/;
  my @new_frames = ();
  my $frame;
  print "adapting format (flags:"
        .(! $scale_int && ! $scale_float ? " none" : "")
        .($scale_int && ! $scale_float ? " scale-int" : "")
        .($scale_float ? " scale-float" : "")
        .(($scale_int || $scale_float) && $scale_aspect ? " change-aspect" : "")
        .") ";
  foreach (@{$old_frames}) {
    push (@new_frames, scale_frame ($_, $new_height, $new_width, $scale_int, $scale_float, $scale_aspect));
    print ".";
  }
  print "\n";
  return ($delays, [@new_frames]);
}

# scale a frame
# usage $frame = scale_frame ($frame, $new_height, $new_width, $scale_int, $scale_float, $scale_aspect)
sub scale_frame {
  my $old_frame = shift;
  my $new_height = bound (shift, 1, 1000);
  my $new_width = bound (shift, 1, 1000);
  my $scale_int = shift;
  my $scale_float = shift;
  my $scale_aspect = shift;
  my $old_height = bound (@{$old_frame}+0, 1, 1000);
  my $old_width = 1;
  $old_width = max ($old_width, @{$_}+0) foreach (@{$old_frame});
  $old_width = bound ($old_width, 1, 1000);

  my $scale_hor = $new_width / $old_width;
  my $scale_ver = $new_height / $old_height;
  if (! $scale_aspect) {
    $scale_hor = $scale_ver = min ($scale_hor, $scale_ver);
  }
  if (! $scale_float) {
    $scale_hor = bound (int ($scale_hor), 1, 1000);
    $scale_ver = bound (int ($scale_ver), 1, 1000);
    if (! $scale_int) {
      $scale_hor = $scale_ver = 1;
    }
  }

  my $old_pic_height = $old_height;
  my $pic_height = $old_pic_height * $scale_ver;
  my $fill_top = int (($new_height - $pic_height) / 2);
  my $skip_top = 0;
  if( $fill_top < 0 ) { # down-scaling not possible due to flags
    $scale_ver = 1; # this should already be the case
    $old_pic_height = $new_height;
    $skip_top = int (($old_height - $new_height) / 2);
    $fill_top = 0;
  }
  my $old_pic_width = $old_width;
  my $pic_width = $old_pic_width * $scale_hor;
  my $fill_left = int (($new_width - $pic_width) / 2);
  my $skip_left = 0;
  if( $fill_left < 0 ) { # down-scaling not possible due to flags
    $scale_hor = 1; # this should already be the case
    $old_pic_width = $new_width;
    $skip_left = int (($old_width - $new_width) / 2);
    $fill_left = 0;
  }

  my @new_frame = ();

  # integer-factor scaling
  if (! $scale_float) {
    my @line;
    my ($x, $y, $xx, $yy, $old_line, $val);
    @new_frame = ();
    @line = ();
    for ($x = 0; $x < $new_width; $x++) { push (@line, 0); }
    for ($y = 0; $y < $fill_top; $y++) { push (@new_frame, [@line]); } 
    for ($y = 0; $y < $old_pic_height; $y++ ) {
      @line = ();
      for ($x = 0; $x < $fill_left; $x++) { push (@line, 0); }
      $old_line = @{$old_frame}[$y+$skip_top];
      for ($x = 0; $x < $old_pic_width; $x++) {
        $val = @{$old_line}[$x+$skip_left];
        for ($xx = 0; $xx < $scale_hor; $xx++) { push (@line, $val); }
      }
      for ($x = 0; $x < $new_width - $fill_left - $pic_width; $x++) { push (@line, 0); }
      for ($yy = 0; $yy < $scale_ver; $yy++) { push (@new_frame, [@line]); }
    }
    @line = ();
    for ($x = 0; $x < $new_width; $x++) { push (@line, 0); }
    for ($y = 0; $y < $new_height - $fill_top - $pic_height; $y++) { push (@new_frame, [@line]); }
  }

  # floating-point-factor-scaling
  else {
    my @line;
    my ($nx, $ny, $ox, $oy, $ox1, $oy1, $oxi, $oyi, $ox1i, $oy1i, $x, $y, $val);
    @new_frame = ();
    for ($ny = 0; $ny < $new_height; $ny++) {
      @line = ();
      for ($nx = 0; $nx < $new_width; $nx++) {
        $oy = ($ny - $fill_top) / $scale_ver + $skip_top; # sub-pixel exact range in old picture
        $ox = ($nx - $fill_left) / $scale_hor + $skip_left;
        $oy1 = ($ny - $fill_top + 1) / $scale_ver + $skip_top - 0.000001;
        $ox1 = ($nx - $fill_left + 1) / $scale_hor + $skip_left - 0.000001;
        if ($oy < 0 || $ox < 0 || $oy1 >= $old_height || $ox1 >= $old_width) { # out of old picture
          push (@line, 0);
        }
        else {
          $oyi = int ($oy);
          $oxi = int ($ox);
          $oy1i = int ($oy1);
          $ox1i = int ($ox1);
          if ($oyi == $oy1i) {
            if ($oxi == $ox1i) { # one source pixel
              $val = @{@{$old_frame}[$oyi]}[$oxi];
            }
            else { # one line of source pixels
              $val = @{@{$old_frame}[$oyi]}[$oxi] * (1 - $ox + $oxi)
                   + @{@{$old_frame}[$oyi]}[$ox1i] * ($ox1 - $ox1i);
              for ($x = $oxi + 1; $x < $ox1i; $x++ ) {
                $val += @{@{$old_frame}[$oyi]}[$x];
              }
              $val /= $ox1 - $ox;
            }
          }
          else { # one column of source pixels
            if ($oxi == $ox1i) {
              $val = @{@{$old_frame}[$oyi]}[$oxi] * (1 - $oy + $oyi)
                   + @{@{$old_frame}[$oy1i]}[$oxi] * ($oy1 - $oy1i);
              for ($y = $oyi + 1; $y < $oy1i; $y++ ) {
                $val += @{@{$old_frame}[$y]}[$oxi];
              }
              $val /= $oy1 - $oy;
            }
            else { # rectangle of source pixels
              $val = @{@{$old_frame}[$oyi]}[$oxi] * (1 - $oy + $oyi) * (1 - $ox + $oxi)
                   + @{@{$old_frame}[$oyi]}[$ox1i] * (1 - $oy + $oyi) * ($ox1 - $ox1i)
                   + @{@{$old_frame}[$oy1i]}[$oxi] * ($oy1 - $oy1i) * (1 - $ox + $oxi)
                   + @{@{$old_frame}[$oy1i]}[$ox1i] * ($oy1 - $oy1i) * ($ox1 - $ox1i);
              for ($y = $oyi + 1; $y < $oy1i; $y++ ) {
                $val += @{@{$old_frame}[$y]}[$oxi] * (1 - $ox + $oxi)
                      + @{@{$old_frame}[$y]}[$ox1i] * ($ox1 - $ox1i);
              }
              for ($x = $oxi + 1; $x < $ox1i; $x++ ) {
                $val += @{@{$old_frame}[$oyi]}[$x] * (1 - $oy + $oyi)
                     + @{@{$old_frame}[$oy1i]}[$x] * ($oy1 - $oy1i);
              }
              for ($y = $oyi + 1; $y < $oy1i; $y++ ) {
                for ($x = $oxi + 1; $x < $ox1i; $x++ ) {
                  $val += @{@{$old_frame}[$y]}[$x];
                }
              }
              $val /= ($oy1 - $oy) * ($ox1 - $ox);
            }
          }
          push (@line, int ($val + 0.5));
        }
      }
      push (@new_frame, [@line]);
    }
  }

  return [@new_frame];
}



# print a movie
# usage: print_movie ($delays, $frames)
sub print_movie {
  my $delays = shift;
  my $frames = shift;
  my $i;
  for ($i = 0; $i < @{$delays}; $i++) {
    print "frame ".$i.": ".@{$delays}[$i]." ms\n";
    print_frame (@{$frames}[$i]);
    print "\n";
  }
}

# print a frame
# usage: print_frame ($frame)
sub print_frame {
  my $frame = shift;
  foreach (@{$frame}) {
    foreach (@{$_}) {
      printf substr (' .,:;-+/*%@#8&$', $_, 1);
    }
    print "\n";
  }
}



# encode a movie
# usage: $commands = enc_movie ($movie_no, $delays, $frames, $flags, $actions)
# $flags: 'c': allow compression
#         'C': allow differential compression
sub enc_movie {
  my $movie_no = shift;
  my $delays = shift;
  my $frames = shift;
  my $flags = shift;
  my $actions = shift;
  my $compr = $flags =~ /c/;
  my $compr_diff = $flags =~ /C/;
  my $frame_commands;
  my @display_commands;
  my @delay_commands;
  my @commands;
  my $i;
  print "encoding movie (flags:"
        .(! $compr && ! $compr_diff ? " none" : "")
        .($compr && ! $compr_diff ? " compression" : "")
        .($compr_diff ? " differential compression" : "")
        .") ";
  @commands = ( [(6, "jump", "MOVIE".($movie_no+1)."_START")],
                @{gen_action_code ($movie_no, $actions)},
                [(0, "label", "MOVIE".($movie_no+1)."_START")] );
  for ($i = 0; $i < @{$delays}; $i++) {
    $frame_commands = enc_frame (@{$frames}[$i], $i > 0 ? @{$frames}[$i-1] : undef, $compr, $i > 0 ? $compr_diff : 0);
    @display_commands = (pack ('C', 0xC1));
    @delay_commands = ( [(3, "delay_start", @{$delays}[$i])],
                        [(6, "call", "MOVIE".($movie_no+1)."_ACTIONS")] );
    push (@commands, @{$frame_commands}, @display_commands, @delay_commands);
    print ".";
  }
  print "\n";
  return [@commands];
}

# generate action code
# usage: $commands = gen_action_code ($movie_no, $actions)
sub gen_action_code {
  my $movie_no = shift;
  my $actions = shift;
  my @keys_commands = ();
  my @snes0_commands = ();
  my @snes1_commands = ();
  # get masks for actions
  my ($start_keys, $start_snes0, $start_snes1) = gen_key_masks ($actions, $ACTION_START);
  my ($prev_keys, $prev_snes0, $prev_snes1) = gen_key_masks ($actions, $ACTION_PREV);
  my ($again_keys, $again_snes0, $again_snes1) = gen_key_masks ($actions, $ACTION_AGAIN);
  my ($next_keys, $next_snes0, $next_snes1) = gen_key_masks ($actions, $ACTION_NEXT);
  my ($end_keys, $end_snes0, $end_snes1) = gen_key_masks ($actions, $ACTION_END);
  # generate code for keys and snes keys
  if ($start_keys | $prev_keys | $again_keys | $next_keys | $end_keys != 0) {
    push (@keys_commands, [(4, "keys")]);
    push (@keys_commands, [(22, "pop6_jump_if_keys", $start_keys, "MOVIES_START")]) if ($start_keys != 0);
    push (@keys_commands, [(22, "pop6_jump_if_keys", $prev_keys, "MOVIE".($movie_no+1-1)."_START")]) if ($prev_keys != 0);
    push (@keys_commands, [(22, "pop6_jump_if_keys", $again_keys, "MOVIE".($movie_no+1)."_START")]) if ($again_keys != 0);
    push (@keys_commands, [(22, "pop6_jump_if_keys", $next_keys, "MOVIE".($movie_no+1+1)."_START")]) if ($next_keys != 0);
    push (@keys_commands, [(22, "pop6_jump_if_keys", $end_keys, "MOVIES_END")]) if ($end_keys != 0);
  }
  if ($start_snes0 | $prev_snes0 | $again_snes0 | $next_snes0 | $end_snes0 != 0) {
    push (@snes0_commands, [(7, "snes0")]);
    push (@snes0_commands, [(26, "pop6_jump_if_snes", $start_snes0, "MOVIES_START")]) if ($start_snes0 != 0);
    push (@snes0_commands, [(26, "pop6_jump_if_snes", $prev_snes0, "MOVIE".($movie_no+1-1)."_START")]) if ($prev_snes0 != 0);
    push (@snes0_commands, [(26, "pop6_jump_if_snes", $again_snes0, "MOVIE".($movie_no+1)."_START")]) if ($again_snes0 != 0);
    push (@snes0_commands, [(26, "pop6_jump_if_snes", $next_snes0, "MOVIE".($movie_no+1+1)."_START")]) if ($next_snes0 != 0);
    push (@snes0_commands, [(26, "pop6_jump_if_snes", $end_snes0, "MOVIES_END")]) if ($end_snes0 != 0);
  }
  if ($start_snes1 | $prev_snes1 | $again_snes1 | $next_snes1 | $end_snes1 != 0) {
    push (@snes1_commands, [(7, "snes1")]);
    push (@snes1_commands, [(26, "pop6_jump_if_snes", $start_snes1, "MOVIES_START")]) if ($start_snes1 != 0);
    push (@snes1_commands, [(26, "pop6_jump_if_snes", $prev_snes1, "MOVIE".($movie_no+1-1)."_START")]) if ($prev_snes1 != 0);
    push (@snes1_commands, [(26, "pop6_jump_if_snes", $again_snes1, "MOVIE".($movie_no+1)."_START")]) if ($again_snes1 != 0);
    push (@snes1_commands, [(26, "pop6_jump_if_snes", $next_snes1, "MOVIE".($movie_no+1+1)."_START")]) if ($next_snes1 != 0);
    push (@snes1_commands, [(26, "pop6_jump_if_snes", $end_snes1, "MOVIES_END")]) if ($end_snes1 != 0);
  }
  # put generated code into delay loop
  return [( [(0, "label", "MOVIE".($movie_no+1)."_ACTIONS")],
            @keys_commands, @snes0_commands, @snes1_commands,
            [(12, "jump_if_delay", "MOVIE".($movie_no+1)."_ACTIONS")],
            pack ('C', 0x87) )];
}

# generate key masks
# usage: ($keys, $snes0, $snes1) = gen_key_masks ($actions, $action)
sub gen_key_masks {
  my $actions = shift;
  my $action = shift;
  my $key;
  my ($keys, $snes0, $snes1) = (0, 0, 0);
  for ($key = 0; $key < 8; $key++) { $keys |= 1 << $key if (@{$actions}[$key] == $action); }
  for ($key = 0; $key < 16; $key++) { $snes0 |= 1 << $key if (@{$actions}[$key+8] == $action); }
  for ($key = 0; $key < 16; $key++) { $snes1 |= 1 << $key if (@{$actions}[$key+8+16] == $action); }
  return ($keys, $snes0, $snes1);
}

# encode a frame
# usage: $commands = enc_frame ($frame, undef, $compr, $compr_diff)
#        $commands = enc_frame ($frame, $last_frame, $compr, $compr_diff)
sub enc_frame {
  my $frame = shift;
  my $last_frame = shift;
  my $compr = shift;
  my $compr_diff = shift;
  my ($commands, $length, $cmds, $len);
  # basic encoding always works
  ($commands, $length) = enc_frame_basic ($frame);
  # try some "better" encoding techniques
  if ($compr || $compr_diff) {
    ($cmds, $len) = enc_frame_bw ($frame);
    ($commands, $length) = ($cmds, $len) if ($len < $length);
  }
  if ($compr_diff) {
    ($cmds, $len) = enc_frame_diff ($frame, $last_frame);
    ($commands, $length) = ($cmds, $len) if ($len < $length);
  }
  # return best encoding
  return $commands;
}

# encode a frame using basic encoding technique
# usage: ($commands, $length) = enc_frame_basic ($frame)
sub enc_frame_basic {
  my $frame = shift;
  my $length = 0;
  my @commands = ();
  my $cmd;
  my ($yy, $y, $x);
  # 4 quarters
  for ($yy = 0; $yy < 4; $yy++) {
    $cmd = pack ('C', 0xA5) . pack ('C', $yy);
    for ($y = $yy * 5; $y < $yy * 5 + 5; $y++) {
      for ($x = 0; $x < 26; $x += 2) {
        $cmd .= pack ('C', @{@{$frame}[$y]}[$x] << 4 | @{@{$frame}[$y]}[$x+1]);
      }
    }
    $length += length ($cmd);
    push (@commands, $cmd);
  }
  return ([@commands], $length)
}

# encode a frame using black/white encoding technique
# usage: ($commands, $length) = enc_frame_bw ($frame)
sub enc_frame_bw {
  my $frame = shift;
  my @colors = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
  my ($y, $x, $c, $cnt, $color, $i);
  my $length = 0;
  my @commands = ();
  my $cmd;
  # check which colors are in frame
  for ($y = 0; $y < 20; $y++) {
    for ($x = 0; $x < 26; $x++) {
      $colors[@{@{$frame}[$y]}[$x]] = 1;
    }
  }
  # count colors that are not black or white
  $cnt = 0;
  for ($c = 1; $c < 15; $c++) {
    if ($colors[$c]) {
      $cnt++;
      $color = $c; # remember that color
    }
  }
  # black/white encoding makes no sense if more than one other color,
  # so return a very high length in this case
  return ([()], 1000000) if ($cnt > 1);
  # black/white pixels
  $cmd = pack ('C', 0xAC);
  $c = 0;
  $i = 0;
  for ($y = 0; $y < 20; $y++) {
    for ($x = 0; $x < 26; $x++) {
      $c <<= 1;
      $c |= 1 if (@{@{$frame}[$y]}[$x] == 0x0F);
      $i++;
      if ($i >= 8) {
        $cmd .= pack ('C', $c);
        $i = 0;
      }
    }
  }
  $length += length ($cmd);
  push (@commands, $cmd);
  # pixels with color $color
  if ($cnt > 0) {
    $cmd = pack ('C', 0xAF) . pack ('C', $color);
    $c = 0;
    $i = 0;
    for ($y = 0; $y < 20; $y++) {
      for ($x = 0; $x < 26; $x++) {
        $c <<= 1;
        $c |= 1 if (@{@{$frame}[$y]}[$x] == $color);
        $i++;
        if ($i >= 8) {
          $cmd .= pack ('C', $c);
          $i = 0;
        }
      }
    }
    $length += length ($cmd);
    push (@commands, $cmd);
  }
  return ([@commands], $length)
}

# encode a frame using differential encoding technique
# usage: ($commands, $length) = enc_frame_diff ($frame, $last_frame)
sub enc_frame_diff {
  my $frame = shift;
  my $last_frame = shift;
  my @lines;
  my ($line_cnt, $pix_cnt);
  my $length = 0;
  my @commands = ();
  my $cmd;
  my ($yy, $y, $x, $l);
  # 4 quarters
  for ($yy = 0; $yy < 4; $yy++) {
    # check which lines changed
    $line_cnt = 0;
    for ($l = 0, $y = $yy * 5; $l < 5; $l++, $y++) {
      $lines[$l] = 0;
      for ($x = 0; $x < 26; $x++) {
        $lines[$l] = 1 if (@{@{$frame}[$y]}[$x] != @{@{$last_frame}[$y]}[$x]);
      }
      $line_cnt++ if ($lines[$l]);
    }
    # too much lines changed in this quarter
    if ($line_cnt > 3) {
      # encode quarter
      $cmd = pack ('C', 0xA5) . pack ('C', $yy);
      for ($y = $yy * 5; $y < $yy * 5 + 5; $y++) {
        for ($x = 0; $x < 26; $x += 2) {
          $cmd .= pack ('C', @{@{$frame}[$y]}[$x] << 4 | @{@{$frame}[$y]}[$x+1]);
        }
      }
      $length += length ($cmd);
      push (@commands, $cmd);
    }
    # only a few lines changed in this quarter
    else {
      # encode only changed lines
      for ($l = 0, $y = $yy * 5; $l < 5; $l++, $y++) {
        if ($lines[$l]) {
          # count number of changed pixels in this line
          $pix_cnt = 0;
          for ($x = 0; $x < 26; $x++) {
            $pix_cnt++ if (@{@{$frame}[$y]}[$x] != @{@{$last_frame}[$y]}[$x]);
          }
          # too much pixels changed in this line
          if ($pix_cnt > 3) {
            # encode line
            $cmd = pack ('C', 0xA1) . pack ('C', $y);
            for ($x = 0; $x < 26; $x += 2) {
              $cmd .= pack ('C', @{@{$frame}[$y]}[$x] << 4 | @{@{$frame}[$y]}[$x+1]);
            }
            $length += length ($cmd);
            push (@commands, $cmd);
          }
          # only a few pixels changed in this line
          else {
            # encode only changed pixels
            for ($x = 0; $x < 26; $x++) {
              if (@{@{$frame}[$y]}[$x] != @{@{$last_frame}[$y]}[$x]) {
                $cmd = pack ('C', 0x99) . pack ('C', $x) . pack ('C', $y) . pack ('C', @{@{$frame}[$y]}[$x]);
                $length += length ($cmd);
                push (@commands, $cmd);
              }
            }
          }
        }
      } # for ($l ...
    }
  } # for ($yy ...
  return ([@commands], $length)
}



# arrange commands into sectors if not done yet
# usage: $flat_commands = flatten_commands ($movies, $movie_cnt, $loop, $movie_names)
sub flatten_commands {
  my $movies = shift;
  my $movie_cnt = shift;
  my $loop = shift;
  my $movie_names = shift;
  my $m;
  my @flat_commands = ();
  push (@flat_commands, [(0, "label", "MOVIES_START")]);
  push (@flat_commands, [(3, "delay_start", 0)]);
  push (@flat_commands, [(0, "label", "MOVIE".($movie_cnt+1)."_START")]);
  for ($m = 0; $m < @{$movies}; $m++) {
    push (@flat_commands, [(0, "movie_name", @{$movie_names}[$m])]);
    push (@flat_commands, [(0, "label", "MOVIE0_START")]) if ($m == @{$movies} - 1);
    push (@flat_commands, @{@{$movies}[$m]});
  }
  push (@flat_commands, [(6, "jump", "MOVIES_START")]) if ($loop);
  push (@flat_commands, [(0, "label", "MOVIES_END")]);
  push (@flat_commands, [(3, "delay_start", 0)]);
  return [@flat_commands];
}



# write amasm
# usage: write_amasm ($commands, $filename)
sub write_amasm {
  my $commands = shift;
  my $filename = shift;
  my @labels = ();
  my ($keys, $snes) = (0, 0);
  my ($ofs, $sec, $cmd, $dot);
  my @data;
  my @command_names = ();
  my ($cmd_no, $cmd_name);
  my $i;
  my $label_no = 0;
  open (F, ">".$filename) or die "write_amasm: can't open output file \"$filename\": $!\n";
  print "writing amasm \"$filename\" ";
  # get used labels
  $dot = 0;
  foreach $cmd (@{$commands}) {
    # array command
    if (ref $cmd) {
      if (@{$cmd}[1] eq "jump" || @{$cmd}[1] eq "call" || @{$cmd}[1] eq "jump_if_delay") {
        $i = 0;
        foreach (@labels) { $i = 1 if (@{$cmd}[2] eq $_); }
        push (@labels, @{$cmd}[2]) if (! $i);
      }
      elsif (@{$cmd}[1] eq "pop6_jump_if_keys" || @{$cmd}[1] eq "pop6_jump_if_snes") {
        $i = 0;
        foreach (@labels) { $i = 1 if (@{$cmd}[3] eq $_); }
        push (@labels, @{$cmd}[3]) if (! $i);
      }
      elsif (@{$cmd}[1] eq "keys") {
        $keys = 1;
      }
      elsif (@{$cmd}[1] eq "snes0" || @{$cmd}[1] eq "snes1") {
        $snes = 1;
      }
    } # if (ref $cmd)
    print "." if ($dot == 0);
    $dot++;
    $dot = 0 if ($dot >= 16);
  } # foreach $cmd ...
  # write amasm
  print F "//this file was automatically generated by b2am\n\n";
  print F "\$DUMMY:         NEXT_REG;\n" if ($keys || $snes);
  print F "\$DELAY:         NEXT_REG;\n";
  print F "\$KEYSL:         NEXT_REG;\n" if ($keys || $snes);
  print F "\$KEYSH:         NEXT_REG;\n" if ($snes);
  print F "\n";
  $command_names[0x87] = "RET";
  $command_names[0x99] = "WR_PIX";
  $command_names[0xA1] = "WR_PIX_LINE";
  $command_names[0xA5] = "WR_PIX_QUART";
  $command_names[0xAC] = "WR_PIX_BW";
  $command_names[0xAF] = "WR_PIX_SEL";
  $command_names[0xC1] = "DISPLAY_CP";
  $ofs = 0;
  $sec = 0;
  foreach $cmd (@{$commands}) {
    # array command
    if (ref $cmd) {
      if ($ofs + @{$cmd}[0] >= 0x200) {
        print F "\n                NEXT_SEC;\n\n";
        $ofs = 0;
        $sec++;
      }
      if (@{$cmd}[1] eq "label") {
        $i = 0;
        foreach (@labels) { $i = 1 if (@{$cmd}[2] eq $_); }
        print F "@".@{$cmd}[2].":\n" if ($i);
      }
      elsif (@{$cmd}[1] eq "jump") {
        print F "                JMP_OFS_SEC     @".@{$cmd}[2].";\n";
      }
      elsif (@{$cmd}[1] eq "call") {
        print F "                CALL            @".@{$cmd}[2].";\n";
      }
      elsif (@{$cmd}[1] eq "delay_start") {
        print F "                DELAY_START     ".@{$cmd}[2].";\n";
      }
      elsif (@{$cmd}[1] eq "jump_if_delay") {
        print F "                DELAY_CHECK     \$DELAY;\n";
        print F "                JMP_EQ          \@LABEL".$label_no.", \$DELAY, 0;\n";
        print F "                JMP_OFS_SEC     @".@{$cmd}[2].";\n";
        print F "\@LABEL".$label_no.":\n";
        $label_no++;
      }
      elsif (@{$cmd}[1] eq "keys") {
        print F "                GET_KEYS        \$DUMMY, \$KEYSL, \$DUMMY;\n";
      }
      elsif (@{$cmd}[1] eq "snes0") {
        print F "                GET_SNES0       \$DUMMY, \$DUMMY, \$KEYSH, \$KEYSL, \$DUMMY, \$DUMMY;\n";
      }
      elsif (@{$cmd}[1] eq "snes1") {
        print F "                GET_SNES1       \$DUMMY, \$DUMMY, \$KEYSH, \$KEYSL, \$DUMMY, \$DUMMY;\n";
      }
      elsif (@{$cmd}[1] eq "pop6_jump_if_keys") {
        print F "                JMP_AND_Z       \@LABEL".$label_no.", \$KEYSL, ".(@{$cmd}[2]&0xFF).";\n";
        print F "                POP             \$DUMMY;\n";
        print F "                POP             \$DUMMY;\n";
        print F "                POP             \$DUMMY;\n";
        print F "                POP             \$DUMMY;\n";
        print F "                POP             \$DUMMY;\n";
        print F "                POP             \$DUMMY;\n";
        print F "                JMP_OFS_SEC     @".@{$cmd}[3].";\n";
        print F "\@LABEL".$label_no.":\n";
        $label_no++;
      }
      elsif (@{$cmd}[1] eq "pop6_jump_if_snes") {
        print F "                JMP_AND_NZ      \@LABEL".$label_no.", \$KEYSH, ".((@{$cmd}[2]>>8)&0xFF).";\n";
        print F "                JMP_AND_Z       \@LABEL".($label_no+1).", \$KEYSL, ".(@{$cmd}[2]&0xFF).";\n";
        print F "\@LABEL".$label_no.":\n";
        print F "                POP             \$DUMMY;\n";
        print F "                POP             \$DUMMY;\n";
        print F "                POP             \$DUMMY;\n";
        print F "                POP             \$DUMMY;\n";
        print F "                POP             \$DUMMY;\n";
        print F "                POP             \$DUMMY;\n";
        print F "                JMP_OFS_SEC     @".@{$cmd}[3].";\n";
        print F "\@LABEL".($label_no+1).":\n";
        $label_no += 2;
      }
      elsif (@{$cmd}[1] eq "movie_name") {
        print F "\n//movie file: ".@{$cmd}[2]."\n\n";
      }
      else {
        die "write_amasm: internal error (command=\"".@{$cmd}[1]."\")\n";
      }
      $ofs += @{$cmd}[0];
    } # if (ref $cmd)
    # binary command
    else {
      if ($ofs + length ($cmd) >= 0x200) {
        print F "\n                NEXT_SEC;\n\n";
        $ofs = 0;
        $sec++;
      }
      @data = split (undef, $cmd);
      $_ = unpack ('C', $_) foreach (@data);
      $cmd_no = shift (@data);
      $cmd_name = $command_names[$cmd_no];
      if ($cmd_name ne "") {
        print F "                " . $cmd_name;
        if (@data > 0) {
          print F substr("                ", length ($cmd_name));
          print F shift (@data);
          print F ", " . $_ foreach (@data);
          print F ";\n";
        }
        else {
          print F ";\n";
        }
      }
      else {
        die "write_amasm: internal error (command=$cmd_no)\n";
      }
      $ofs += length ($cmd);
    } # if (ref $cmd) ... else
    print "." if ($dot == 0);
    $dot++;
    $dot = 0 if ($dot >= 16);
  } # foreach $cmd ...
  print "\n";
  close F;
}

# write ambin
# usage: write_ambin ($commands, $filename)
sub write_ambin {
  my $commands = shift;
  my $filename = shift;
  my ($reg_dummy, $reg_delay, $reg_keysl, $reg_keysh) = (0, 1, 2, 3);
  my ($ofs, $sec, $cmd, $dot, $jump_ofs, $jump_sec);
  my @labels = ();
  my $i;
  open (F, ">".$filename) or die "write_ambin: can't open output file \"$filename\": $!\n";
  binmode F;
  print F "ARCADEmini 26x20/16 ambin 1.0   ";
  for ($i = 0x20; $i < 0x200; $i++) { print F pack ('C', 0xFF); }
  print "writing ambin \"$filename\" ";
  # get addresses of labels
  $dot = 0;
  foreach $cmd (@{$commands}) {
    # array command
    if (ref $cmd) {
      if ($ofs + @{$cmd}[0] >= 0x200) {
        $ofs = 0;
        $sec++;
      }
      if (@{$cmd}[1] eq "label") {
        push (@labels, [(@{$cmd}[2], $sec, $ofs)]);
      }
      $ofs += @{$cmd}[0];
    } # if (ref $cmd)
    # binary command
    else {
      if ($ofs + length ($cmd) >= 0x200) {
        $ofs = 0;
        $sec++;
      }
      $ofs += length ($cmd);
    } # if (ref $cmd) ... else
    print "." if ($dot == 0);
    $dot++;
    $dot = 0 if ($dot >= 16);
  } # foreach $cmd ...
  # write ambin
  $ofs = 0;
  $sec = 0;
  foreach $cmd (@{$commands}) {
    # array command
    if (ref $cmd) {
      if ($ofs + @{$cmd}[0] >= 0x200) {
        for ( ; $ofs < 0x200; $ofs++) { print F pack ('C', 0x00 ); }
        $ofs = 0;
        $sec++;
      }
      if (@{$cmd}[1] eq "label") {
        # ignore labels
      }
      elsif (@{$cmd}[1] eq "jump") {
        ($jump_sec, $jump_ofs) = (-1, -1);
        foreach (@labels) { ($jump_sec, $jump_ofs) = (@{$_}[1], @{$_}[2]) if (@{$cmd}[2] eq @{$_}[0]); }
        die "write_ambin: internal error (unknown label \"".@{$cmd}[2]."\")\n" if ($jump_ofs < 0);
        $jump_sec -= $sec;
        print F pack ('C', 0x04 | (($jump_ofs >> 8) & 0x01))
              . pack ('C', $jump_ofs & 0xFF)
              . pack ('C', ($jump_sec >> 24) & 0xFF)
              . pack ('C', ($jump_sec >> 16) & 0xFF)
              . pack ('C', ($jump_sec >> 8) & 0xFF)
              . pack ('C', $jump_sec & 0xFF);
      }
      elsif (@{$cmd}[1] eq "call") {
        ($jump_sec, $jump_ofs) = (-1, -1);
        foreach (@labels) { ($jump_sec, $jump_ofs) = (@{$_}[1], @{$_}[2]) if (@{$cmd}[2] eq @{$_}[0]); }
        die "write_ambin: internal error (unknown label \"".@{$cmd}[2]."\")\n" if ($jump_ofs < 0);
        $jump_sec -= $sec;
        print F pack ('C', 0x84 | (($jump_ofs >> 8) & 0x01))
              . pack ('C', $jump_ofs & 0xFF)
              . pack ('C', ($jump_sec >> 24) & 0xFF)
              . pack ('C', ($jump_sec >> 16) & 0xFF)
              . pack ('C', ($jump_sec >> 8) & 0xFF)
              . pack ('C', $jump_sec & 0xFF);
      }
      elsif (@{$cmd}[1] eq "delay_start") {
        print F pack ('C', 0xCB)
              . pack ('C', (@{$cmd}[2] >> 8) & 0xFF)
              . pack ('C', @{$cmd}[2] & 0xFF);
      }
      elsif (@{$cmd}[1] eq "jump_if_delay") {
        print F pack ('C', 0xCC)
              . pack ('C', $reg_delay)
              . pack ('C', 0x0A | ((($ofs + 12) >> 8) & 0x01))
              . pack ('C', ($ofs + 12) & 0xFF)
              . pack ('C', $reg_delay)
              . pack ('C', 0);
        ($jump_sec, $jump_ofs) = (-1, -1);
        foreach (@labels) { ($jump_sec, $jump_ofs) = (@{$_}[1], @{$_}[2]) if (@{$cmd}[2] eq @{$_}[0]); }
        die "write_ambin: internal error (unknown label \"".@{$cmd}[2]."\")\n" if ($jump_ofs < 0);
        $jump_sec -= $sec;
        print F pack ('C', 0x04 | (($jump_ofs >> 8) & 0x01))
              . pack ('C', $jump_ofs & 0xFF)
              . pack ('C', ($jump_sec >> 24) & 0xFF)
              . pack ('C', ($jump_sec >> 16) & 0xFF)
              . pack ('C', ($jump_sec >> 8) & 0xFF)
              . pack ('C', $jump_sec & 0xFF);
      }
      elsif (@{$cmd}[1] eq "keys") {
        print F pack ('C', 0xD0)
              . pack ('C', $reg_dummy)
              . pack ('C', $reg_keysl)
              . pack ('C', $reg_dummy);
      }
      elsif (@{$cmd}[1] eq "snes0") {
        print F pack ('C', 0xD2)
              . pack ('C', $reg_dummy)
              . pack ('C', $reg_dummy)
              . pack ('C', $reg_keysh)
              . pack ('C', $reg_keysl)
              . pack ('C', $reg_dummy)
              . pack ('C', $reg_dummy);
      }
      elsif (@{$cmd}[1] eq "snes1") {
        print F pack ('C', 0xD3)
              . pack ('C', $reg_dummy)
              . pack ('C', $reg_dummy)
              . pack ('C', $reg_keysh)
              . pack ('C', $reg_keysl)
              . pack ('C', $reg_dummy)
              . pack ('C', $reg_dummy);
      }
      elsif (@{$cmd}[1] eq "pop6_jump_if_keys") {
        print F pack ('C', 0x32 | ((($ofs + 22) >> 8) & 0x01))
              . pack ('C', ($ofs + 22) & 0xFF)
              . pack ('C', $reg_keysl)
              . pack ('C', @{$cmd}[2] & 0xFF)
              . pack ('C', 0x82)
              . pack ('C', $reg_dummy)
              . pack ('C', 0x82)
              . pack ('C', $reg_dummy)
              . pack ('C', 0x82)
              . pack ('C', $reg_dummy)
              . pack ('C', 0x82)
              . pack ('C', $reg_dummy)
              . pack ('C', 0x82)
              . pack ('C', $reg_dummy)
              . pack ('C', 0x82)
              . pack ('C', $reg_dummy);
        ($jump_sec, $jump_ofs) = (-1, -1);
        foreach (@labels) { ($jump_sec, $jump_ofs) = (@{$_}[1], @{$_}[2]) if (@{$cmd}[3] eq @{$_}[0]); }
        die "write_ambin: internal error (unknown label \"".@{$cmd}[3]."\")\n" if ($jump_ofs < 0);
        $jump_sec -= $sec;
        print F pack ('C', 0x04 | (($jump_ofs >> 8) & 0x01))
              . pack ('C', $jump_ofs & 0xFF)
              . pack ('C', ($jump_sec >> 24) & 0xFF)
              . pack ('C', ($jump_sec >> 16) & 0xFF)
              . pack ('C', ($jump_sec >> 8) & 0xFF)
              . pack ('C', $jump_sec & 0xFF);
      }
      elsif (@{$cmd}[1] eq "pop6_jump_if_snes") {
        print F pack ('C', 0x36 | ((($ofs + 8) >> 8) & 0x01))
              . pack ('C', ($ofs + 8) & 0xFF)
              . pack ('C', $reg_keysh)
              . pack ('C', (@{$cmd}[2] >> 8) & 0xFF)
              . pack ('C', 0x32 | ((($ofs + 26) >> 8) & 0x01))
              . pack ('C', ($ofs + 26) & 0xFF)
              . pack ('C', $reg_keysl)
              . pack ('C', @{$cmd}[2] & 0xFF)
              . pack ('C', 0x82)
              . pack ('C', $reg_dummy)
              . pack ('C', 0x82)
              . pack ('C', $reg_dummy)
              . pack ('C', 0x82)
              . pack ('C', $reg_dummy)
              . pack ('C', 0x82)
              . pack ('C', $reg_dummy)
              . pack ('C', 0x82)
              . pack ('C', $reg_dummy)
              . pack ('C', 0x82)
              . pack ('C', $reg_dummy);
        ($jump_sec, $jump_ofs) = (-1, -1);
        foreach (@labels) { ($jump_sec, $jump_ofs) = (@{$_}[1], @{$_}[2]) if (@{$cmd}[3] eq @{$_}[0]); }
        die "write_ambin: internal error (unknown label \"".@{$cmd}[3]."\")\n" if ($jump_ofs < 0);
        $jump_sec -= $sec;
        print F pack ('C', 0x04 | (($jump_ofs >> 8) & 0x01))
              . pack ('C', $jump_ofs & 0xFF)
              . pack ('C', ($jump_sec >> 24) & 0xFF)
              . pack ('C', ($jump_sec >> 16) & 0xFF)
              . pack ('C', ($jump_sec >> 8) & 0xFF)
              . pack ('C', $jump_sec & 0xFF);
      }
      elsif (@{$cmd}[1] eq "movie_name") {
        # ignore movie name
      }
      else {
        die "write_ambin: internal error (command=\"".@{$cmd}[1]."\")\n";
      }
      $ofs += @{$cmd}[0];
    } # if (ref $cmd)
    # binary command
    else {
      if ($ofs + length ($cmd) >= 0x200) {
        for ( ; $ofs < 0x200; $ofs++) { print F pack ('C', 0x00 ); }
        $ofs = 0;
        $sec++;
      }
      print F $cmd;
      $ofs += length ($cmd);
    } # if (ref $cmd) ... else
    print "." if ($dot == 0);
    $dot++;
    $dot = 0 if ($dot >= 16);
  } # foreach $cmd ...
  # fill last sector
  for ( ; $ofs < 0x200; $ofs++) { print F pack ('C', 0x00 ); }
  $ofs = 0;
  $sec++;
  # jump to first sector
  print F pack ('C', 0x01)
        . pack ('C', ((-$sec) >> 24) & 0xFF)
        . pack ('C', ((-$sec) >> 16) & 0xFF)
        . pack ('C', ((-$sec) >> 8) & 0xFF)
        . pack ('C', (-$sec) & 0xFF);
  for ( $i = 5; $i < 0x200; $i++) { print F pack ('C', 0x00 ); }
  print "\n";
  close F;
}



# bound a value
# usage: $val = bound ($val, $lower, $upper);
sub bound {
  my ($val, $lower, $upper) = @_;
  return min (max ($val, $lower), $upper);
}

# minimum value
# usage: $min = min ($a, $b, ...);
sub min {
  my $val = shift;
  foreach (@_) {
    $val = $_ if ($_ < $val);
  }
  return $val;
}

# maximum value
# usage: $max = max ($a, $b, ...);
sub max {
  my $val = shift;
  foreach (@_) {
    $val = $_ if ($_ > $val);
  }
  return $val;
}
